// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file MultiGridSolverBase.h */

#ifndef AMDIS_MULTIGRIDSOLVERBASE_H
#define AMDIS_MULTIGRIDSOLVERBASE_H

#include "InterpolRestrictMatrix.h"
#include "LagrangeInterpolRestrict.h"
#include <string>

namespace AMDiS {

  template<typename MatrixType, typename VectorType>
  class MultiGridSolverBase
  {
  public:
    /** \brief
     * Constructor, reads all multigrid parameters from the init file.
     */
    MultiGridSolverBase(const std::string &name);

    /** \brief
     * Destructor
     */
    virtual ~MultiGridSolverBase() {}

    /** \brief
     * Abstract function for solving an equation using multigrid method.
     * This function does not need to be reimplemented by a derived class.
     */
    virtual int solve(MatrixType &a, 
		      VectorType &x,
		      VectorType &b,
		      bool initMG = true,
		      bool exitMG = true);

    /** \brief
     * Implements one multigrid iteration. This function does not need to be
     * reimplemented by a derived class.
     */
    virtual void multiGridIteration(int level);

    virtual bool toleranceReached(double tol) = 0;

    virtual void initMaxLevel() = 0;

    virtual void initMultiGrid(bool createLevels) = 0;

    virtual void exitMultiGrid() = 0;

    virtual void smooth(int level, int steps) = 0;

    virtual void restrict(int level, int coarseLevel) = 0;

    virtual void prolongate(int level, int coarseLevel) = 0;

    virtual void computeResidual(int level, int coarseLevel) = 0;

    virtual void computeRHS(int level, int coarseLevel) = 0;

    virtual void computeCorrection(int level, int coarseLevel) = 0;

    virtual void fixSolution(int level, int coarseLevel) = 0;

    virtual void exactSolve(int level) = 0;

    virtual int coarserLevel(int level) = 0;

    virtual void initFullMultiGrid() {};

    virtual void fmgInterpol(int level) {};

  protected:
    /** \brief
     * Coefficent matrix
     */
    MatrixType *systemMatrix_;

    /** \brief
     * Vector for the computed solution.
     */
    VectorType *solution_;

    /** \brief
     * Right hand side of the equation.
     */
    VectorType *rhs_;

    /** \brief
     * Name of the problem, corresponding to the name in the init file.
     */
    std::string name_;

    /** \brief
     * Tolerance for the computation. The multigrid method finished, if the
     * norm of the residual is less than the tolerance.
     */
    double tolerance_;

    /** \brief
     * Maximum number of multigrid iterations.
     */
    int maxIterations_;

    /** \brief
     * The coarsest multigrid level.
     */
    int minLevel_;

    /** \brief
     * The finest multigrid level.
     */
    int maxLevel_;

    int numCycles_;

     /** \brief
     * Number of smoothing steps on the solution before the recursive multigrid call.
     */
    int preSmoothingSteps_;

    /** \brief
     * Number of smoothing steps on the solution after the recursive multigrid call.
     */
    int postSmoothingSteps_;

    /** \brief
     * If the method of full multigrid should be used, this variable has a non
     * zero value.
     */
    int fmg_;

    /** \brief
     * interpolMatrices[n][m] stores the interpolation matrix for
     * n-dimensional elements and m-dimensional basis functions.
     */
    InterpolRestrictMatrix *interpolMatrices[3][4];

    /** \brief
     * interpolMatrices[n][m] stores the restriction matrix for
     * n-dimensional elements and m-dimensional basis functions.
     */
    InterpolRestrictMatrix *restrictMatrices[3][4];
  };

}

#include "MultiGridSolverBase.hh"
#endif
