#include "AMDiS.h"
#include "ParallelProblem.h"
#include "mpi.h"

using namespace std;
using namespace AMDiS;

// ===========================================================================
// ===== function definitions ================================================
// ===========================================================================

/** \brief
 * Dirichlet boundary function
 */
class G : public AbstractFunction<double, WorldVector<double> >,
	  public TimedObject
{
public:
  MEMORY_MANAGED(G);

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  double operator()(const WorldVector<double>& argX) const {
    WorldVector<double> x = argX;
    int dim = x.getSize();
    for (int i = 0; i < dim; i++) {
      x[i] -= *timePtr;
    }
    return sin(M_PI*(*timePtr)) * exp(-10.0*(x*x));
  }
};

/** \brief
 * RHS function
 */
class F : public AbstractFunction<double, WorldVector<double> >,
	  public TimedObject
{
public:
  MEMORY_MANAGED(F);

  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  double operator()(const WorldVector<double>& argX) const {
    WorldVector<double> x = argX;
    int dim = x.getSize();
    for (int i = 0; i < dim; i++) {
      x[i] -= *timePtr;
    }

    double r2 = (x*x);
    double ux = sin(M_PI * (*timePtr)) * exp(-10.0*r2);
    double ut = M_PI * cos(M_PI*(*timePtr)) * exp(-10.0*r2);
    return = ut -(400.0*r2 - 20.0*dim)*ux;
  }
};

// ===========================================================================
// ===== instationary problem ================================================
// ===========================================================================

/** \brief
 * Instationary problem
 */
class Heat : public ProblemInstatScal
{
public:
  MEMORY_MANAGED(Heat);

  /** \brief
   *  Constructor
   */
  Heat(ProblemScal *heatSpace)
    : ProblemInstatScal("heat", heatSpace)
  {};


  // ===== ProblemInstatBase methods ===================================

  /** \brief
   * set the time in all needed functions!
   */
  void setTime(AdaptInfo *adaptInfo) {
    time = adaptInfo->getTime();
    tau1 = 1.0 / adaptInfo->getTimestep();    
  };

  // ===== initial problem methods =====================================

  /** \brief
   * Used by \ref problemInitial to solve the system of the initial problem
   */
  void solve(AdaptInfo *adaptInfo) 
  {
    problemStat->getMesh()->dofCompress();
    time = adaptInfo->getStartTime();
    problemStat->getSolution()->interpol(boundaryFct);
  };

  /** \brief
   * Used by \ref problemInitial to do error estimation for the initial
   * problem.
   */
  void estimate(AdaptInfo *adaptInfo) 
  {
    double errMax, errSum;

    time = adaptInfo->getStartTime();

    errSum = Error<double>::L2Err(*boundaryFct,
				  *(problemStat->getSolution()), 0, &errMax, false);

    adaptInfo->setEstSum(errSum, 0);
  };

  // ===== setting methods ===============================================

  /** \brief
   * Sets \ref boundaryFct;
   */
  void setBoundaryFct(AbstractFunction<double, WorldVector<double> > *fct) 
  {
    boundaryFct = fct;
  } 

  // ===== getting methods ===============================================


  /** \brief
   * Returns pointer to \ref tau1
   */
  double *getTau1Ptr() 
  { 
    return &tau1; 
  };

  /** \brief
   * Returns pointer to \ref time.
   */
  double *getTimePtr() 
  { 
    return &time; 
  };

  /** \brief
   * Returns \ref boundaryFct;
   */
  AbstractFunction<double, WorldVector<double> > *getBoundaryFct() 
  {
    return boundaryFct;
  };


//   void closeTimestep(AdaptInfo *adaptInfo) {
//     ParallelProblem::writeRankMacroAndValues(problemStat->getSolution(),
// 					     "output/debug_rank", 
// 					     adaptInfo->getTime());
//   };

private:
  /** \brief
   * 1.0 / timestep
   */
  double tau1;

  /** \brief
   * time 
   */
  double time;

  /** \brief
   * Pointer to boundary function. Needed for initial problem.
   */
  AbstractFunction<double, WorldVector<double> > *boundaryFct;
};

// ===========================================================================
// ===== main program ========================================================
// ===========================================================================

int main(int argc, char** argv)
{
  MPI::Init(argc, argv);

  // ===== check for init file =====
  TEST_EXIT(argc >= 2)("usage: heat initfile\n");

  // ===== init parameters =====
  Parameters::init(false, argv[1]);

  // ===== create and init stationary problem =====
  ProblemScal *heatSpace = NEW ProblemScal("heat->space");
  heatSpace->initialize(INIT_ALL);

  // ===== create instationary problem =====
  Heat *heat = new Heat(heatSpace);
  heat->initialize(INIT_ALL);

  // create adapt info
  AdaptInfo *adaptInfo = NEW AdaptInfo("heat->adapt");

  // create initial adapt info
  AdaptInfo *adaptInfoInitial = NEW AdaptInfo("heat->initial->adapt");

  // create parallel problem
  std::vector<DOFVector<double>*> vectors;
  ParallelProblemScal parallelheat("heat->parallel", heatSpace, heat, vectors);

  // create instationary adapt
  AdaptInstationary *adaptInstat =
    NEW AdaptInstationary("heat->adapt",
			  &parallelheat,
			  adaptInfo,
			  &parallelheat,
			  adaptInfoInitial);

  // ===== create boundary functions =====
  G *boundaryFct = NEW G;
  boundaryFct->setTimePtr(heat->getTimePtr());
  heat->setBoundaryFct(boundaryFct);

  heatSpace->addDirichletBC(DIRICHLET, boundaryFct);

  // ===== create rhs functions =====
  int degree = heatSpace->getFESpace()->getBasisFcts()->getDegree();
  F *rhsFct = NEW F(degree);
  rhsFct->setTimePtr(heat->getTimePtr());

  // ===== create operators =====
  double one = 1.0;
  double zero = 0.0;

  // create laplace
  Operator *A = NEW Operator(Operator::MATRIX_OPERATOR | 
			     Operator::VECTOR_OPERATOR,
			     heatSpace->getFESpace());

  A->addSecondOrderTerm(new Laplace_SOT);

  A->setUhOld(heat->getOldSolution());

  heatSpace->addMatrixOperator(A, &one, &one);

  // create zero order operator
  Operator *C = NEW Operator(Operator::MATRIX_OPERATOR | 
			     Operator::VECTOR_OPERATOR,
			     heatSpace->getFESpace());

  C->addZeroOrderTerm(NEW Simple_ZOT);

  C->setUhOld(heat->getOldSolution());

  heatSpace->addMatrixOperator(C, heat->getTau1Ptr(), heat->getTau1Ptr());
  heatSpace->addVectorOperator(C, heat->getTau1Ptr(), heat->getTau1Ptr());

  // create RHS operator
  Operator *F = NEW Operator(Operator::VECTOR_OPERATOR,
			     heatSpace->getFESpace());

  F->addZeroOrderTerm(NEW CoordsAtQP_ZOT(rhsFct));

  heatSpace->addVectorOperator(F);

//   // ===== start adaption loop =====
//   if(MPI::COMM_WORLD.Get_rank() == 0) {
//     bool wait=true; 
//     while(wait) {}
//   }

  parallelheat.initParallelization(adaptInfo);

  adaptInstat->adapt();

  parallelheat.exitParallelization(adaptInfo);

//   if(MPI::COMM_WORLD.Get_rank() == 0)
//     heatSpace->writeFiles(adaptInfo, true);

  MPI::Finalize();
}
